@@ -89,7 +89,6 @@ pub extern crate serde;
 extern crate test;
 
 #[macro_use]
-#[cfg(test)]
 mod macros;
 
 pub mod descriptor;
@@ -122,7 +121,7 @@ pub trait MiniscriptKey:
     /// The associated Hash type with the publicKey
     type Hash: Clone + Eq + Ord + str::FromStr + fmt::Display + fmt::Debug + hash::Hash;
 
-    ///Converts an object to PublicHash
+    /// Converts an object to PublicHash
     fn to_pubkeyhash(&self) -> Self::Hash;
 }
 
@@ -270,8 +269,6 @@ pub enum Error {
     NonMinimalVerify(miniscript::lex::Token),
     /// Push was illegal in some context
     InvalidPush(Vec<u8>),
-    /// PSBT-related error
-    Psbt(psbt::Error),
     /// rust-bitcoin script error
     Script(script::Error),
     /// A `CHECKMULTISIG` opcode was preceded by a number > 20
@@ -370,6 +367,13 @@ impl From<miniscript::context::ScriptContextError> for Error {
     }
 }
 
+#[doc(hidden)]
+impl From<bitcoin::secp256k1::Error> for Error {
+    fn from(e: bitcoin::secp256k1::Error) -> Error {
+        Error::Secp(e)
+    }
+}
+
 fn errstr(s: &str) -> Error {
     Error::Unexpected(s.to_owned())
 }
@@ -378,7 +382,6 @@ impl error::Error for Error {
     fn cause(&self) -> Option<&error::Error> {
         match *self {
             Error::BadPubkey(ref e) => Some(e),
-            Error::Psbt(ref e) => Some(e),
             _ => None,
         }
     }
@@ -399,7 +402,6 @@ impl fmt::Display for Error {
             Error::InvalidOpcode(op) => write!(f, "invalid opcode {}", op),
             Error::NonMinimalVerify(tok) => write!(f, "{} VERIFY", tok),
             Error::InvalidPush(ref push) => write!(f, "invalid push {:?}", push), // TODO hexify this
-            Error::Psbt(ref e) => fmt::Display::fmt(e, f),
             Error::Script(ref e) => fmt::Display::fmt(e, f),
             Error::CmsTooManyKeys(n) => write!(f, "checkmultisig with {} keys", n),
             Error::Unprintable(x) => write!(f, "unprintable character 0x{:02x}", x),
@@ -460,13 +462,6 @@ impl fmt::Display for Error {
     }
 }
 
-#[doc(hidden)]
-impl From<psbt::Error> for Error {
-    fn from(e: psbt::Error) -> Error {
-        Error::Psbt(e)
-    }
-}
-
 #[doc(hidden)]
 #[cfg(feature = "compiler")]
 impl From<policy::compiler::CompilerError> for Error {
@@ -482,6 +477,13 @@ impl From<policy::concrete::PolicyError> for Error {
     }
 }
 
+#[doc(hidden)]
+impl From<descriptor::InterpreterError> for Error {
+    fn from(e: descriptor::InterpreterError) -> Error {
+        Error::InterpreterError(e)
+    }
+}
+
 /// The size of an encoding of a number in Script
 pub fn script_num_size(n: usize) -> usize {
     match n {
