@@ -25,12 +25,23 @@ use bitcoin::hashes::{hash160, ripemd160, sha256, sha256d};
 use bitcoin::{self, secp256k1};
 use {MiniscriptKey, ToPublicKey};
 
+use miniscript::types::extra_props::HEIGHT_TIME_THRESHOLD;
+use Error;
 use ScriptContext;
 use Terminal;
 
 /// Type alias for a signature/hashtype pair
 pub type BitcoinSig = (secp256k1::Signature, bitcoin::SigHashType);
 
+/// Helper function to create BitcoinSig from Rawsig
+/// Useful for downstream when implementing Satisfier.
+/// Returns underlying secp if the Signature is not of correct formart
+pub fn bitcoinsig_from_rawsig(rawsig: &[u8]) -> Result<BitcoinSig, Error> {
+    let (flag, sig) = rawsig.split_last().unwrap();
+    let flag = bitcoin::SigHashType::from_u32(*flag as u32);
+    let sig = secp256k1::Signature::from_der(sig)?;
+    Ok((sig, flag))
+}
 /// Trait describing a lookup table for signatures, hash preimages, etc.
 /// Every method has a default implementation that simply returns `None`
 /// on every query. Users are expected to override the methods that they
@@ -94,7 +105,12 @@ pub struct Older(pub u32);
 
 impl<Pk: MiniscriptKey> Satisfier<Pk> for Older {
     fn check_older(&self, n: u32) -> bool {
-        n <= self.0
+        // if n > self.0; we will be returning false anyways
+        if n < HEIGHT_TIME_THRESHOLD && self.0 >= HEIGHT_TIME_THRESHOLD {
+            false
+        } else {
+            n <= self.0
+        }
     }
 }
 
@@ -104,7 +120,12 @@ pub struct After(pub u32);
 
 impl<Pk: MiniscriptKey> Satisfier<Pk> for After {
     fn check_after(&self, n: u32) -> bool {
-        n <= self.0
+        // if n > self.0; we will be returning false anyways
+        if n < HEIGHT_TIME_THRESHOLD && self.0 >= HEIGHT_TIME_THRESHOLD {
+            false
+        } else {
+            n <= self.0
+        }
     }
 }
 
